# -*- coding: utf-8 -*-

"""
 (c) 2019 - Copyright Red Hat Inc

 Authors:
   Pierre-Yves Chibon <pingou@pingoured.fr>
   Karsten Hopp <karsten@redhat.com>

"""

from __future__ import print_function, unicode_literals

import collections
import logging

import flask

import pagure.utils
from pagure.api import APIERROR, api_login_required, api_method
from pagure.api.utils import _check_token, _get_repo
from pagure.exceptions import PagureException

import pagure_distgit.forms
from pagure_distgit import model

import requests

from sqlalchemy.exc import SQLAlchemyError


_log = logging.getLogger(__name__)

DISTGIT_NS = flask.Blueprint(
    "distgit_ns", __name__, url_prefix="/_dg", template_folder="templates"
)


@DISTGIT_NS.route("/anitya/<namespace>/<repo>", methods=["GET"])
def anitya_get_endpoint(namespace, repo):
    """Returns the current status of the monitoring in anitya of this
    package."""
    repo = flask.g.repo
    output = {"monitoring": "no-monitoring"}
    if repo.anitya:
        output = {"monitoring": repo.anitya.anitya_status}
    return flask.jsonify(output)


@DISTGIT_NS.route("/anitya/<namespace>/<repo>", methods=["POST"])
@api_login_required(acls=["modify_project"])
@api_method
def anitya_patch_endpoint(namespace, repo):
    """Updates the current status of the monitoring in anitya of this
    package."""

    repo = _get_repo(repo, namespace=namespace)
    _check_token(repo, project_token=False)

    is_site_admin = pagure.utils.is_admin()
    admins = [u.username for u in repo.get_project_users("admin")]
    # Only allow the main admin, the admins of the project, and Pagure site
    # admins to modify projects' monitoring, even if the user has the right
    # ACLs on their token
    if (
        flask.g.fas_user.username not in admins
        and flask.g.fas_user.username != repo.user.username
        and not is_site_admin
    ):
        raise pagure.exceptions.APIError(
            401, error_code=APIERROR.EMODIFYPROJECTNOTALLOWED
        )

    form = pagure_distgit.forms.AnityaForm(csrf_enabled=False)
    if form.validate_on_submit():
        try:
            if repo.anitya:
                repo.anitya.anitya_status = form.anitya_status.data
                flask.g.session.add(repo)
            else:
                mapping = model.PagureAnitya(
                    project_id=repo.id, anitya_status=form.anitya_status.data
                )
                flask.g.session.add(mapping)
            flask.g.session.commit()
        except SQLAlchemyError as err:  # pragma: no cover
            flask.g.session.rollback()
            _log.exception(err)
            raise pagure.exceptions.APIError(400, error_code=APIERROR.EDBERROR)
    else:
        raise pagure.exceptions.APIError(
            400, error_code=APIERROR.EINVALIDREQ, errors=form.errors
        )

    return anitya_get_endpoint(namespace, repo.name)


@DISTGIT_NS.route("/actived/<namespace>/<repo>", methods=["GET"])
@api_method
def get_actived_status(namespace, repo):
    """Retrieves the active status of the specified package."""
    _log.info("Received a request to unorphan: %s/%s", namespace, repo)

    repo = _get_repo(repo, namespace=namespace)

    if repo.user.user != "orphan":
        raise pagure.exceptions.APIError(
            401, error_code=APIERROR.EMODIFYPROJECTNOTALLOWED
        )

    # Check if the project is retired in PDC
    active = _is_active_in_pdc(repo.name, repo.namespace)

    output = {"active": active}
    return flask.jsonify(output)


def _is_active_in_pdc(name, namespace):
    """Queries PDC and return whether the project is active on the master
    branch in PDC or not.
    """
    pdc_url = flask.current_app.config.get("PDC_URL")
    if not pdc_url:
        raise pagure.exceptions.APIError(
            500,
            error_code=APIERROR.ENOCODE,
            error="This pagure instance has no PDC_URL configured, please "
            "inform your pagure administrators",
        )
    else:
        pdc_url = "%s/component-branches/" % pdc_url.rstrip("/")

    _log.debug("Based PDC url: %s", pdc_url)

    to_pdc_namespace = {
        "rpms": "rpm",
        "modules": "module",
        "container": "container",
        "flatpaks": "flatpak"
    }
    to_pdc_namespace = (
        flask.current_app.config.get("PDC_NAMESPACES") or to_pdc_namespace
    )

    try:
        pdc_namespace = to_pdc_namespace[namespace]
    except Exception:
        raise pagure.exceptions.APIError(
            500,
            error_code=APIERROR.ENOCODE,
            error="Namespace: %s could not be converted to a PDC namespace"
            % namespace,
        )

    branch = "master"
    if namespace in ["rpms", "container"]:
        branch = "rawhide"
    elif namespace == "flatpaks":
        branch = "stable"

    url = "%s?global_component=%s&name=%s&type=%s" % (
        pdc_url,
        name,
        branch,
        pdc_namespace,
    )

    _log.info("Querying PDC at: %s", url)
    try:
        req = requests.get(url, timeout=(30, 30))
    except requests.RequestException as err:
        raise pagure.exceptions.APIError(
            500,
            error_code=APIERROR.ENOCODE,
            error="An error occured while querying pdc: %s" % err,
        )

    try:
        data = req.json()
    except Exception:
        raise pagure.exceptions.APIError(
            500,
            error_code=APIERROR.ENOCODE,
            error="The output of %s could not be converted to JSON" % req.url,
        )

    _log.info(
        "%s/%s is active: %s", namespace, name, data["results"][0]["active"]
    )
    return data["results"][0]["active"] is True


@DISTGIT_NS.route("/orphan/<namespace>/<repo>", methods=["GET"])
@api_method
def orphan_get_endpoint(namespace, repo):
    """
    Orphan state
    ------------
    Returns the current orphan state of package with reason.

    ::

        GET /_dg/orphan/<namespace>/<repo>

    Sample response
    ^^^^^^^^^^^^^^^

    ::

        {
          "orphan": true,
          "reason": "Lack of Time",
          "reason_info": "Personal issues"
        }
    """
    repo = _get_repo(repo, namespace=namespace)

    output = {"orphan": False, "reason": "", "reason_info": ""}

    if repo.user.user == "orphan":
        output["orphan"] = True
        if repo.orphan_reason:
            output["reason"] = repo.orphan_reason.reason
            output["reason_info"] = repo.orphan_reason.reason_info
    return flask.jsonify(output)


@DISTGIT_NS.route("/orphan/<namespace>/<repo>", methods=["POST"])
@api_login_required(acls=["modify_project"])
@api_method
def orphan_endpoint(namespace, repo):
    """
    Orphan
    ------
    Orphan the package.

    ::

        POST /_dg/orphan/<namespace>/<repo>

    Input
    ^^^^^

    +-----------------------+---------+--------------+------------------------+
    | Key                   | Type    | Optionality  | Description            |
    +=======================+=========+==============+========================+
    | ``orphan_reason``     | string  | Optional     | | The reason to orphan |
    |                       |         |              |   the package.         |
    +-----------------------+---------+--------------+------------------------+
    | ``orphan_reason_info``| string  | Optional     | | Additional info for  |
    |                       |         |              |   provided reason.     |
    +-----------------------+---------+--------------+------------------------+

    Sample response
    ^^^^^^^^^^^^^^^

    ::

        {
          "orphan": true,
          "reason": "Other",
          "reason_info": "Voices in my head tell me to stop"
        }
    """
    _log.info("Received a request to orphan: %s/%s", namespace, repo)

    repo = _get_repo(repo, namespace=namespace)
    _check_token(repo, project_token=False)

    user_obj = pagure.lib.query.get_user(
        flask.g.session, flask.g.fas_user.username
    )
    if not user_obj:
        raise pagure.exceptions.APIError(404, error_code=APIERROR.ENOUSER)

    if repo.user.user == "orphan":
        raise pagure.exceptions.APIError(
            401, error_code=APIERROR.EMODIFYPROJECTNOTALLOWED
        )

    if repo.user.user != flask.g.fas_user.username \
            and not pagure.utils.is_admin():
        raise pagure.exceptions.APIError(
            401, error_code=APIERROR.EMODIFYPROJECTNOTALLOWED
        )

    try:
        orphan_user_obj = pagure.lib.query.get_user(flask.g.session, "orphan")
    except PagureException:
        _log.exception("Error when retrieving orphan user")
        raise pagure.exceptions.APIError(400, error_code=APIERROR.EDBERROR)

    form = pagure_distgit.forms.OrphanReasonForm(csrf_enabled=False)
    if form.validate_on_submit():
        try:
            reason = model.PagureOrphanReason(
                project_id=repo.id,
                reason=form.orphan_reason.data,
                reason_info=form.orphan_reason_info.data,
            )
            flask.g.session.add(reason)
            pagure.lib.query.set_project_owner(
                flask.g.session, repo, orphan_user_obj
            )
            if user_obj in repo.users:
                pagure.lib.query.remove_user_of_project(
                    flask.g.session, user_obj, repo, user_obj.user
                )
            pagure.lib.query.update_watch_status(
                flask.g.session, repo, user_obj.username, "-1"
            )
            flask.g.session.commit()
        except SQLAlchemyError:  # pragma: no cover
            flask.g.session.rollback()
            _log.exception("Error when orphaning project")
            raise pagure.exceptions.APIError(400, error_code=APIERROR.EDBERROR)
    else:
        raise pagure.exceptions.APIError(
            400, error_code=APIERROR.EINVALIDREQ, errors=form.errors
        )

    pagure.lib.notify.log(
        repo,
        topic="project.orphan",
        msg={
            "project": repo.to_json(public=True),
            "agent": user_obj.username,
            "reason": repo.orphan_reason.reason,
            "reason_info": repo.orphan_reason.reason_info,
        },
    )

    return orphan_get_endpoint(namespace, repo.name)


@DISTGIT_NS.route("/take_orphan/<namespace>/<repo>", methods=["POST"])
@api_login_required(acls=["modify_project"])
@api_method
def take_orphan_endpoint(namespace, repo):
    """Updates the current point of contact of orphan packages."""
    _log.info("Received a request to unorphan: %s/%s", namespace, repo)

    repo = _get_repo(repo, namespace=namespace)
    _check_token(repo, project_token=False)

    user_obj = pagure.lib.query.get_user(
        flask.g.session, flask.g.fas_user.username
    )
    if not user_obj:
        raise pagure.exceptions.APIError(404, error_code=APIERROR.ENOUSER)

    if repo.user.user != "orphan":
        raise pagure.exceptions.APIError(
            401, error_code=APIERROR.EMODIFYPROJECTNOTALLOWED
        )

    user_grps = set(user_obj.groups)
    req_grps = set(["packager"])
    if not user_grps.intersection(req_grps):
        raise pagure.exceptions.APIError(
            403,
            error_code=APIERROR.ENOTHIGHENOUGH,
            errors="You must be a packager to adopt a package.",
        )

    # Check if the project is retired in PDC
    if not _is_active_in_pdc(repo.name, repo.namespace):
        raise pagure.exceptions.APIError(
            400,
            error_code=APIERROR.EINVALIDREQ,
            errors="This project has been retired and cannot be unorphaned "
            "here, please open a releng ticket for it.",
        )

    try:
        repo.user = user_obj
        if repo.orphan_reason:
            reason = repo.orphan_reason
            flask.g.session.delete(reason)
        flask.g.session.add(repo)
        flask.g.session.commit()
    except SQLAlchemyError as err:  # pragma: no cover
        flask.g.session.rollback()
        _log.exception(err)
        raise pagure.exceptions.APIError(400, error_code=APIERROR.EDBERROR)

    pagure.lib.notify.log(
        repo,
        topic="project.adopt",
        msg=dict(project=repo.to_json(public=True), agent=user_obj.username),
    )

    output = {"point_of_contact": repo.user.user}

    return flask.jsonify(output)


def query_oraculum_versions(package):
    """
    Receives package versions from oraculum
    Returns triplet: fedora-releases, updates-for-package, F%number-of-current-rawhide
    """

    oraculum_base_url = "https://packager-dashboard.fedoraproject.org/api/"
    # easier debugging when running directly outside of flask context
    try:
        oraculum_url = flask.current_app.config.get("ORACULUM_URL", oraculum_base_url)
    except RuntimeError:
        oraculum_url = oraculum_base_url

    if not oraculum_url:
        raise pagure.exceptions.APIError(
            500,
            error_code=APIERROR.ENOCODE,
            error="This pagure instance has no ORACULUM_URL configured, please "
            "inform your pagure administrators",
        )

    _log.debug("Based ORACULUM url: %s", oraculum_url)
    req_versions = requests.get("%sv1/package_versions/%s" % (oraculum_url, package))
    req_releases = requests.get("%sv1/releases" % oraculum_url)

    if not req_releases.ok:
        raise pagure.exceptions.APIError(
                400,
                error_code=APIERROR.EINVALIDREQ,
                errors=["Could not call oraculum to query current Fedora Releases."],
            )

    if not req_versions.ok:
        raise pagure.exceptions.APIError(
                400,
                error_code=APIERROR.EINVALIDREQ,
                errors=["Could not call oraculum to query current package versions."],
            )

    versions_resp = req_versions.json()
    releases_resp = req_releases.json()

    # Releases mapping ({"FXX": "Fedora XX", "EPEL X": "Fedora EPEL X"})
    releases = {}
    releases["ELN"] = "Fedora ELN"
    for epel_release in releases_resp["epel"]:
        releases["EPEL-%d" % epel_release] = "Fedora EPEL %d" % epel_release
    for release in releases_resp["fedora"]["values"]:
        releases["F%d" % release] = "Fedora %d" % release

    rawhide_num = max(releases_resp["fedora"]["values"])
    rawhide = "F" + str(rawhide_num)

    # Rename fedora releases to what pagure's dist-git api uses now
    # (Fedora XX => FXX; Fedora Rawhide => FXX, EPEL X => EPEL-X)
    versions = {}

    for release in list(versions_resp.keys()):
        if release in ["Fedora ELN", "Fedora Rawhide"]:
            continue
        if release.startswith("Fedora"):
            versions[release.replace("Fedora ", "F")] = versions_resp[release]
        elif release.startswith("EPEL"):
            versions[release.replace(" ", "-")] = versions_resp[release]

    # And finally Rawhide and ELN
    versions[rawhide] = versions_resp["Fedora Rawhide"]
    versions["ELN"] = versions_resp["Fedora ELN"]

    return releases, versions, rawhide

def query_bodhi_versions(repo):
    """
    Receives package versions from bodhi
    Returns triplet: fedora-releases, updates-for-package, F%number-of-current-rawhide
    """
    bodhi_base_url = "https://bodhi.fedoraproject.org/"

    # Retrieves the active releases from bodhi
    releases_url = "%s/releases/" % bodhi_base_url
    releases = {}
    page = 1
    pages = 1
    while True:
        url = "%s?page=%s" % (releases_url, page)
        _log.info("Calling bodhi at: %s", url)
        req = requests.get(url)
        if not req.ok:
            raise pagure.exceptions.APIError(
                400,
                error_code=APIERROR.EINVALIDREQ,
                errors=["Could not call bodhi"],
            )

        data = req.json()
        for release in data.get("releases", []):
            if any(
                word in release["long_name"]
                for word in ["Flatpaks", "Containers", "Modular"]
            ):
                continue
            if release["state"] in ["pending", "current"]:
                releases[release["name"]] = release["long_name"]

        pages = data["pages"]
        page = data["page"]
        if page == pages:
            break
        page += 1

    # Set Rawhide as the highest-sorted Fedora release
    rawhide_number = 0
    for release in releases:
        # Skip EL / EPEL / ELN releases
        if not release.startswith("F"):
            continue
        # Skip Container, Flatpak, and Module releases
        try:
            base_release = int(release.lstrip("F").rstrip("CFM"))
        except ValueError:
            continue
        if base_release > rawhide_number:
            rawhide_number = base_release
    rawhide = "F" + str(rawhide_number)

    # Retrieves the updates of that package in bodhi
    update_url = "%s/updates?packages=%s" % (
        bodhi_base_url.rstrip("/"),
        repo,
    )
    updates = collections.defaultdict(dict)

    _log.info("Calling bodhi at: %s", update_url)
    req = requests.get(update_url)
    if not req.ok:
        raise pagure.exceptions.APIError(
            400,
            error_code=APIERROR.EINVALIDREQ,
            errors=["Could not call bodhi"],
        )

    data = req.json()

    # To be consistent with oraculum
    for release in releases:
        updates[release] = {"pending": None, "stable": None, "testing": None}

    for update in data["updates"]:
        if update["release"]["name"] not in releases:
            continue
        if update["status"] == "obsolete":
            # To be consistent with oraculum
            continue
        name = update["title"]
        for build in update["builds"]:
            if repo in build["nvr"]:
                name = build["nvr"]
                break

        if not updates[update["release"]["name"]].get(update["status"]):
            updates[update["release"]["name"]][update["status"]] = name

    return releases, updates, rawhide

@DISTGIT_NS.route("/bodhi_updates/<namespace>/<repo>", methods=["GET"])
@api_method
def bodhi_updates_endpoint(namespace, repo):
    """Retrieves the current updates in bodhi for the specified package."""
    if namespace not in ["rpms"]:
        raise pagure.exceptions.APIError(
            400,
            error_code=APIERROR.EINVALIDREQ,
            errors=["Namespace not supported"],
        )

    _log.info(
        "Received request for the bodhi updates of: %s/%s", namespace, repo
    )

    repo = _get_repo(repo, namespace=namespace)
    html = pagure.utils.is_true(flask.request.args.get("html", False))

    try:
        releases, updates, rawhide = query_oraculum_versions(repo.name)
    except pagure.exceptions.APIError:
        releases, updates, rawhide = query_bodhi_versions(repo.name)

    if html:
        html_output = (
            '<table class="table table-bordered table-striped">\n'
            '<tr><th scope="col">Release</th><th scope="col">Stable version'
            '</th><th scope="col">Version in testing</th></tr>\n'
        )
        for release in sorted(releases, reverse=True):
            row = (
                "<tr><td scope='row' class='text-align-left'>{release}</td>"
                "<td>{stable}</td><td>{testing}</td></tr>"
            )
            stable = ""
            if updates.get(release, {}).get("stable"):
                stable = (
                    "<a href='https://koji.fedoraproject.org/koji/search?"
                    "terms={0}&type=build&match=glob'>{0}</a>".format(
                        updates[release].get("stable")
                    )
                )
            else:
                # This makes things a little slower but is needed for builds
                # that do not go through bodhi, for example, via releng-managed
                # side tags (like for Fedora mass rebuilds), which bypass bodhi.
                if release == rawhide:
                    mdapi_release = "rawhide"
                else:
                    mdapi_release = release.lower()
                mdapi_url = "https://mdapi.fedoraproject.org/%s/srcpkg/%s" % (
                    mdapi_release,
                    repo.name,
                )
                _log.info("Calling mdapi at: %s", mdapi_url)
                req = requests.get(mdapi_url)
                if req.ok:
                    _log.info("mdapi returned ok")
                    data = req.json()
                    build = "%s-%s-%s" % (
                        data["basename"],
                        data["version"],
                        data["release"],
                    )
                    stable = (
                        "<a href='https://koji.fedoraproject.org/koji/search?"
                        "terms={0}&type=build&match=glob'>{0}</a>".format(
                            build
                        )
                    )
            testing = ""
            if updates.get(release, {}).get("testing"):
                testing = (
                    "<a href='https://koji.fedoraproject.org/koji/search?"
                    "terms={0}&type=build&match=glob'>{0}</a>".format(
                        updates[release].get("testing")
                    )
                )
            row = row.format(
                release=releases[release], stable=stable, testing=testing
            )
            html_output += row
        html_output += "</table>"
        return html_output
    else:
        return flask.jsonify({"releases": releases, "updates": updates})


@DISTGIT_NS.route("/bzoverrides/<namespace>/<repo>", methods=["GET"])
@api_method
def bzoverride_get_endpoint(repo, namespace):
    """Returns the current default assignee(s) of this package.
    Defaults to the repo user if unset.
    """
    repo = _get_repo(repo, namespace=namespace)
    output = {
        "fedora_assignee": repo.user.username,
        "epel_assignee": repo.user.username,
    }
    if repo.bzoverride:
        if repo.bzoverride.fedora_assignee:
            output["fedora_assignee"] = repo.bzoverride.fedora_assignee
        if repo.bzoverride.epel_assignee:
            output["epel_assignee"] = repo.bzoverride.epel_assignee
    return flask.jsonify(output)


@DISTGIT_NS.route("/bzoverrides/<namespace>/<repo>", methods=["POST"])
@api_method
@api_login_required(acls=["modify_project"])
def bzoverride_patch_endpoint(repo, namespace):
    """Updates the default assignees of this package."""

    def _validate_input(inputname):
        """ Validate if the input is either an username or a group name. """
        valid = False
        if inputname.startswith("@"):
            group = pagure.lib.query.search_groups(
                flask.g.session, group_name=inputname[1:]
            )
            if group:
                valid = True

        else:
            user_obj = pagure.lib.query.search_user(
                flask.g.session, username=inputname
            )
            if user_obj:
                valid = True

        return valid

    repo = _get_repo(repo, namespace=namespace)

    is_site_admin = pagure.utils.is_admin()
    # Only allow the main admin and Pagure site admins to modify projects'
    # monitoring, even if the user has the right ACLs on their token
    if flask.g.fas_user.username != repo.user.username and not is_site_admin:
        raise pagure.exceptions.APIError(
            401, error_code=APIERROR.EMODIFYPROJECTNOTALLOWED
        )

    form = pagure_distgit.forms.BZOverrideForm(csrf_enabled=False)
    if form.validate_on_submit():
        fedora_assignee = None
        if form.fedora_assignee.data:
            fedora_assignee = form.fedora_assignee.data.strip() or None
            if fedora_assignee and not _validate_input(fedora_assignee):
                raise pagure.exceptions.APIError(
                    400,
                    error_code=APIERROR.EINVALIDREQ,
                    errors=["Invalid user or group name as fedora_assignee"],
                )

        epel_assignee = None
        if form.epel_assignee.data:
            epel_assignee = form.epel_assignee.data.strip() or None
            if epel_assignee and not _validate_input(epel_assignee):
                raise pagure.exceptions.APIError(
                    400,
                    error_code=APIERROR.EINVALIDREQ,
                    errors=["Invalid user or group name as epel_assignee"],
                )

        try:
            if repo.bzoverride:
                if fedora_assignee is None and epel_assignee is None:
                    flask.g.session.delete(repo.bzoverride)
                else:
                    repo.bzoverride.fedora_assignee = fedora_assignee
                    repo.bzoverride.epel_assignee = epel_assignee
                    flask.g.session.add(repo)
            elif fedora_assignee is not None or epel_assignee is not None:
                mapping = model.PagureBZOverride(
                    project_id=repo.id,
                    epel_assignee=epel_assignee,
                    fedora_assignee=fedora_assignee,
                )
                flask.g.session.add(mapping)
            flask.g.session.commit()
        except SQLAlchemyError as err:  # pragma: no cover
            flask.g.session.rollback()
            _log.exception(err)
            raise pagure.exceptions.APIError(400, error_code=APIERROR.EDBERROR)
    else:
        raise pagure.exceptions.APIError(
            400, error_code=APIERROR.EINVALIDREQ, errors=form.errors
        )

    return bzoverride_get_endpoint(repo=repo.name, namespace=namespace)
